<?php

/**
 * PHPDevShell is a RAD Framework aimed at developing administrative applications.
 *
 * @package PHPDevShell
 * @link http://www.phpdevshell.org
 * @copyright Copyright (C) 2007 Jason Schoeman, All rights reserved.
 * @license GNU/LGPL, see readme/licensed_under_lgpl or http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
 *
 * Copyright notice: See readme/notice
 * By using PHPDevShell you agree to notice and license, if you dont agree to terms you are not allowed to use PHPDevShell.
 */

/**
 * Build array from get url.
 * 
 * TODO: it's probably faster to use PHP build-in function (array_merge...)
 *
 * @param array $myGET
 * @param array $includeInGet
 * @param array $excludeFromGet
 * @return array
 * 
 * @date 20110809 (v1.0.1) (greg) fixed a typo with $includeInGet
 * @author greg
 */
function PU_BuildGETArray(array $myGET, $includeInGet = null, $excludeFromGet = null)
{
	if (!is_null($includeInGet)) {
		if (!is_array($includeInGet)) $includeInGet = array($includeInGet);
		foreach ($includeInGet as $index => $value)
			$myGET[$index] = $value;
	}
	if (!is_null($excludeFromGet)) {
		if (!is_array($excludeFromGet)) $excludeFromGet = array($excludeFromGet);
		foreach ($excludeFromGet as $param)
			unset($myGET[$param]);
	}
	return $myGET;
}

/**
 * Creates a (string) url to be used with GET, including encoding
 *
 * @param array $myGET
 * @param string $glue
 * @return string
 */
function PU_BuildGETString(array $myGET, $glue = '&amp;')
{
	$url = '';
	if (count($myGET)) {
		$params = array();
		foreach ($myGET as $index => $value)
			$params[] .= rawurlencode($index) . '=' . rawurlencode($value);
		$url = '?' . implode($glue, $params);
	}
	return $url;
}

/**
 * Build GET part of a url
 *
 * @param $includeInGet	(optional) array of pairs: parameters to add as GET in the url
 * @param $excludeFromGet (optional) array of strings: parameters to remove from GET in the url
 * @return string the whole parameter part of the url (including '?') ; maybe empty if there are no parameters
 */
function PU_BuildGET($includeInGet = null, $excludeFromGet = null, $glue = '&amp;')
{
	return PU_BuildGETString(PU_BuildGETArray($_GET, $includeInGet, $excludeFromGet), $glue);
}

/**
 * Build a xml-style attributes string based on an array
 *
 * @version 1.1
 * @date 20091203
 *  @date 20110203 (v1.1) (greg) added the glue parameter
 * @param $attributes array, the attribute array to compile
 * @param $glue string, a piece of string to insert between the values
 * @return string
 */
function PU_BuildAttrString(array $attributes = null, $glue = '')
{
	$result = '';
	if (is_array($attributes))
		foreach ($attributes as $key => $value) {
			if ($result && $glue) $result .= $glue;
			$result .= " $key=\"$value\"";
		}
	return $result;
}

/**
 * Builds a parsed url.
 *
 * @param array $p
 * @return string
 */
function PU_buildParsedURL($p)
{
	if (!is_array($p)) return $p;

	if (empty($p['scheme'])) $p['scheme'] = 'http';
	if (empty($p['host'])) $p['host'] = $_SERVER["HTTP_HOST"];
	if (empty($p['port'])) $p['port'] = ''; else $p['port'] = ':' . $p['port'];
	if (empty($p['user'])) $p['user'] = '';
	if (empty($p['pass'])) $p['pass'] = '';
	if (empty($p['path'])) $p['path'] = $_SERVER["PHP_SELF"];
	if (empty($p['query'])) $p['query'] = '';
	if (empty($p['fragment'])) $p['fragment'] = ''; else
			$p['fragment'] = '#' . $p['fragment'];

	$auth = ($p['user'] || $p['pass']) ? $p['user'] . ':' . $p['pass'] . '@' : '';

	if ($p['query'] && ('?' != substr($p['query'], 0, 1)))
			$p['query'] = '?' . $p['query'];

	if ('/' == substr($p['path'], 0, 1))
			$url = $p['scheme'] . '://' . $auth . $p['host'] . $p['port'] . $p['path'] . $p['query'] . $p['fragment'];
	else $url = $p['path'] . $p['query'] . $p['fragment'];

	return $url;
}

/**
 * Build a url with GET parameters
 *
 * @param string|array $target (optional) string: the target script url (current script if missing)
 * @param array $includeInGet	(optional) array of pairs: parameters to add as GET in the url
 * @param array $excludeFromGet (optional) array of strings: parameters to remove from GET in the url
 * @return string the built url
 *
 * @version 1.1
 * @author greg
 *
 * @date 20100930 (v1.1) (greg) $target parameter can now be an array resulting from php's parse_url function
 */
function PU_BuildURL($target = null, $includeInGet = null, $excludeFromGet = null, $glue = '&amp;')
{
	if (is_null($target)) $target = $_SERVER["REQUEST_URI"];
	if (!is_array($target)) $target = parse_url($target);

	if (empty($target['query'])) $tarGET = $_GET;
	else {
		parse_str($target['query'], $tarGET);
		$tarGET = array_merge($_GET, $tarGET);
	}
	$myGET = PU_BuildGETArray($tarGET, $includeInGet, $excludeFromGet);
	$target['query'] = PU_BuildGETString($myGET, $glue);
	$target = PU_buildParsedURL($target);
	return $target;
}

/**
 * Build a html link (A+HREF html tag) with label and url and GET parameters
 *
 * @version 1.0.1
 * @date 20091203: added $attrs parameter
 * @param $label string: the text of the link
 * @param $includeInGet (optional) array of pairs: parameters to add as GET in the url
 * @param $excludeFromGet (optional) array of strings: parameters to remove from GET in the url
 * @param $target (optional) string: the target script url (current script if missing)
 * @return string the complete html link
 */
function PU_BuildHREF($label, $includeInGet = null, $excludeFromGet = null, $target = null, array $attrs = null)
{
	$url = PU_BuildURL($target, $includeInGet, $excludeFromGet);
	$href = '<a href="' . $url . '">' . $label . '</a>';
	return $href;
}

/**
 * Clean a string from possibly harmful chars
 *
 * These are removed: single and double quotes, backslashes, optionnaly html tags (everything between < and >)
 *
 * A cleaned string should be safe to include in an html output
 *
 * @param $string the string to clean
 * @param $clean_htlm if true, HTML tags are deleted too
 *
 * @return string
 */
function PU_CleanString($string, $clean_htlm = false)
{
	$string = preg_replace('/["\'\\\\]/', '', $string);
	if ($clean_htlm) $string = preg_replace('/<.+>/', '', $string);
	return $string;
}

/**
 * Convert a string to UTF8 (default) or to HTML
 *
 * @param $string the string to convert
 * @param $htmlize if true the string is converted to HTML, otherwise to UTF8
 *
 * @return string
 */
function PU_MakeString($string, $htmlize = false)
{
	if (!empty($string)) {
		$from = mb_detect_encoding($string, "UTF-8, ISO-8859-1, ISO-8859-15", true);
		$to = ($htmlize ? 'HTML-ENTITIES' : 'UTF-8');
		$string = mb_convert_encoding($string, $to, $from);
	}
	return $string;
}

/**
 * Search for array values inside array and returns key.
 *
 * @param array $needle
 * @param array $haystack
 * @return mixed
 */
function PU_ArraySearch($needle, $haystack)
{
	if (empty($needle) || empty($haystack)) {
		return false;
	}

	foreach ($haystack as $key => $value) {
		$exists = 0;
		foreach ($needle as $nkey => $nvalue) {
			if (!empty($value[$nkey]) && $value[$nkey] == $nvalue) {
				$exists = 1;
			} else {
				$exists = 0;
			}
		}
		if ($exists) return $key;
	}

	return false;
}

/**
 * Create gettext functions.
 */
if (function_exists('gettext')) {

	/**
	 * Wrapper for $core->__() method.
	 * Converts text to use gettext PO system. Does the same as $core->__();
	 *
	 * @param string $say_what The string required to output or convert.
	 * @param string $domain Override textdomain that should be looked under for this text string.
	 * @return string Will return converted string or same string if not available.
	 * @author Jason Schoeman
	 */
	function __($gettext, $domain = false)
	{
		// Do we have a textdomain override value?
		if (empty($domain)) {
			// Normal gettext in current domain.
			return gettext("$gettext");
		} else {
			// Lets handle text inside a different domain.
			return dgettext("$domain", "$gettext");
		}
	}

	/**
	 * Specifically meant for core translation domain.
	 *
	 * @param string $say_what The string required to output or convert.
	 * @return string
	 */
	function ___($gettext)
	{
		return dgettext('core.lang', "$gettext");
	}

	/**
	 * This function echos the returning text.
	 *
	 * @param text $text
	 */
	function _e($text)
	{
		echo gettext($text);
	}

	/**
	 * This function echos the returning text inside a domain.
	 *
	 * @param text $text
	 */
	function __e($text, $domain)
	{
		echo dgettext($domain, $text);
	}
} else {

	/**
	 * Specifically meant for core translation domain.
	 *
	 * @param string $say_what The string required to output or convert.
	 * @return string
	 */
	function ___($gettext)
	{
		return dgettext('core.lang', "$gettext");
	}

	// Create dummy functions.
	function gettext($text)
	{
		return $text;
	}

	function dgettext($domain, $text)
	{
		return $text;
	}

	function _($text)
	{
		return $text;
	}

	function __($gettext, $domain = false)
	{
		return $gettext;
	}

	function _e($text)
	{
		echo $text;
	}

	function __e($text)
	{
		echo $text;
	}

	function textdomain ($textdomain)
	{
		return '';
	}
}

/**
 * Outputs an array in html
 *
 * A slightly better version of print_r()
 *
 * @version 1.2
 * @author greg
 *
 * @date 20100825 (greg) (v1.1) updated to deal with associative arrays
 * @date 20110211 (greg) (v1.2) added $htmlize parameter
 *
 * @param array $a
 * @param string $title
 * @param boolean $htmlize (default to false) if true html is escaped to be displayed as source
 */
function PU_dumpArray($a, $title = null, $htmlize = false)
{
	if (!is_array($a)) $a = array($a);
	$s = '';
	if ($title) $s .= "<p>$title</p>";

	if (count($a) == 0) $s .= '<p>Empty array</p>';
	else {
		$s .= '<ul>';
		foreach ($a as $k => $e) {
			if ($htmlize) $e = htmlentities($e);
			$s .= "<li><i>[ $k ]&nbsp;=&gt;</i>&nbsp;$e</li>";
		}
		$s .= '</ul>';
	}
	return $s;
}

/**
 * Get rid of all buffer, optionaly flushing (i.e. writing to the browser)
 * Default behavior is to ignore.
 *
 * @param boolean $flush do we flush or ignore?
 */
function PU_cleanBuffers($flush = false)
{
	try {
		for ($loop = ob_get_level(); $loop; $loop--) {
			$flush ? ob_end_flush() : ob_end_clean();
		}
	} catch (Exception $e) {
		// only mask the exception and prevent it from bubbling
		$a = 0;
	} catch (ErrorException $e) {
		// only mask the exception and prevent it from bubbling
		$a = 0;
	} catch (PHPDS_fatalError $e) {
		// only mask the exception and prevent it from bubbling
		$a = 0;
	}
}

/**
 * Add a header if and only if headers have not been sent yet
 * 
 * @param string $header the header string to add
 * @return nothing
 * 
 * @version 1.0
 * @since 3.0.6
 * @date 20110809 (v1.0) (greg) added
 * @author greg <greg@phpdevshell.org>
 */
function PU_silentHeader($header)
{
	if (!headers_sent()) {
		header($header);
	}
}

/**
 * Determines if the current request has been made by some kind of ajax call (i.e. XMLHttpRequest)
 * 
 * @param boolean $json set to true if you want to force the request's result as json
 * @return boolean
 * 
 * @version 1.0.1
 * @date 20110809 (v1.0.1) (greg) use PU_silentHeader to prevent unit tests from failing
 * @author greg <greg@phpdevshell.org>
 */
function PU_isAJAX($json = false)
{
	$ajax = isset($_SERVER["HTTP_X_REQUESTED_WITH"]) && ($_SERVER["HTTP_X_REQUESTED_WITH"] == 'XMLHttpRequest');
	if ($ajax && $json) {
		PU_cleanBuffers();
		PU_silentHeader('Content-Type: application/json');
	}
	return $ajax;
}

/**
 * Checks for a json context and if so, outputs data
 *
 * @author greg <greg@phpdevshell.org>
 * @version 1.1
 * @since v3.0.1
 * @date 20110309 (v1.0) (greg) added
 * @date 20110316 (v1.1) (greg) returns instead of printing (for buffering)
 *
 * @param $data mixed, the data to be encoded and sent
 * @param $force boolean, (optionnal) do we pretend it's json context even if it's not?
 * @return boolean, false if it's not JSON, or the encoded data
 */
function PU_isJSON($data, $force = false)
{
	$json = $force || (isset($_SERVER["HTTP_X_REQUESTED_TYPE"]) && ($_SERVER["HTTP_X_REQUESTED_TYPE"] == 'json'));
	if ($json && PU_isAJAX(true)) {
		return json_encode($data);
	}
	return false;
}

/**
 * OBSOLETE don't use
 *
 * @param $data
 */
function PU_exitToAJAX($data)
{
	PU_isAJAX(true);
	print json_encode($data);
	exit;
}

/**
 * Get rid of null values inside an array
 *
 * All values which are null in the array are remove, shortening the array
 *
 * @param array $a the array to compact
 */
function PU_array_compact(array $a)
{
	array_walk($a, create_function('&$v,$k,$a', 'if (is_null($a[$k])) unset($a[$k]);'), $a);
	return $a;
}







/**
 * version of sprintf for cases where named arguments are desired (python syntax)
 *
 * with sprintf: sprintf('second: %2$s ; first: %1$s', '1st', '2nd');
 *
 * with sprintfn: sprintfn('second: %(second)s ; first: %(first)s', array(
 *  'first' => '1st',
 *  'second'=> '2nd'
 * ));
 * @author nate at frickenate dot com
 * @param string $format sprintf format string, with any number of named arguments
 * @param array $args array of [ 'arg_name' => 'arg value', ... ] replacements to be made
 * @return string|false result of sprintf call, or bool false on error
 */
function PU_sprintfn($format, array $args = array())
{
	try {
		// map of argument names to their corresponding sprintf numeric argument value
		$arg_nums = array_slice(array_flip(array_keys(array(0 => 0) + $args)), 1);

		// find the next named argument. each search starts at the end of the previous replacement.
		for ($pos = 0; preg_match('/(?<=%)\(([a-zA-Z_]\w*)\)/', $format, $match, PREG_OFFSET_CAPTURE, $pos);) {
			$arg_pos = $match[0][1];
			$arg_len = strlen($match[0][0]);
			$arg_key = $match[1][0];

			// programmer did not supply a value for the named argument found in the format string
			if (!array_key_exists($arg_key, $arg_nums)) {
				throw new PHPDS_Exception("Missing named argument '${arg_key}'");
			}

			// replace the named argument with the corresponding numeric one
			$format = substr_replace($format, $replace = $arg_nums[$arg_key] . '$', $arg_pos, $arg_len);
			$pos = $arg_pos + strlen($replace); // skip to end of replacement for next iteration
		}

		$result = vsprintf($format, array_values($args));
		return $result;
	} catch (Exception $e) {
		throw new PHPDS_sprintfnException($format, $args, $e);
	}
}


/**
 * Create a html string of <options> from an associative array
 *
 * @version 1.0.1
 * @date 20101021 (v1.0.1) (greg) added checks on $a ; removed htmlentities
 * @author greg
 *
 * @param array $a
 * @param string|array $selected which key(s) should be marked as "selected" (optional)
 * @return string the html to display
 */
function PU_buildHTMLoptions($a, $selected = null)
{
	$options = '';
	if (is_array($a) && (count($a) > 0))
			foreach ($a as $key => $value) {
			$options .= '<option value="' . htmlspecialchars($key) . '"';
			if (($key == $selected) || (is_array($selected) && in_array($key, $selected)))
					$options .=' selected ';
			$options .= '>' . ($value) . '</option>';
		}
	return $options;
}

/**
 * Add an include path to check in for classes.
 * @param string $path
 */
function PU_addIncludePath($path)
{
	if (!empty($path) && is_string($path)) {
		set_include_path(get_include_path() . PATH_SEPARATOR . $path);
	}
}

/**
 * A class to deal with tree-structured data (such as groups)
 *
 * Usage:
 * $tree = new PU_tree
 * $tree->add(1, 0); // add a root node
 * $tree->add(2, 1, 'leaf'); // add a named leaf to the root node
 * $tree->climb(); // study the tree YOU HAVE TO DO THAT BEFORE ACTUALLY USING THE TREE
 *
 * @author greg
 * @date 20100514
 * @version	1.0
 *
 */
class PU_tree
{
	/**
	 * Associative array of the nodes (element ref => element, usually a label)
	 *
	 * @var array
	 */
	protected $elements = array(0);
	/**
	 * Associative array: for each node (by ref), what are the nodes upper in the tree
	 *
	 * @var array
	 */
	protected $ascendants = array();
	/**
	 * Associative array: for each node (by ref), what are the nodes lower in the tree
	 *
	 * @var array
	 */
	protected $descendants = array();

	/**
	 * Add an element to the tree. When all elements are added, you MUST call climb()
	 *
	 * @param mixed $leaf the new element ref
	 * @param mixed $node the element onto this new element is stuck
	 * @param mixed $label an optional label to display
	 * @return this
	 */
	public function add($leaf, $node, $label = '')
	{
		$this->elements[$leaf] = $label;
		$this->descendants[$node][] = $leaf;

		return $this;
	}

	/**
	 * Climb the tree in order to fill the descendant array.
	 * Don't call it with parameter
	 *
	 * @param unknown_type $branch
	 * @return this
	 */
	public function climb($branch = 0)
	{
		if (! empty($this->descendants[$branch]) && is_array($this->descendants[$branch])) {
			foreach ($this->descendants[$branch] as $leaf) {
				$this->ascendants[$leaf] = (empty($branch)) ? array() : $this->ascendants[$branch];
				if ($branch) $this->ascendants[$leaf][] = $branch;
				$this->climb($leaf);
			}
		}
		return $this;
	}

	/**
	 * Returns the ascendants of the given node, either as array or as a string for sql
	 *
	 * @param $node the node which ascendants are asked
	 * @param $as_array	boolean, do you want an array (true) or a string (false)
	 * @return array or string
	 */
	public function ascendants($node, $as_array = false)
	{
		if ($as_array)
			return (isset($this->ascendants[$node]) ? $this->ascendants[$node] : array());
		else
			return (isset($this->ascendants[$node]) ? implode(',', $this->ascendants[$node]) : '');
	}

	/**
	 * Returns the descendants of the given node, either as array or as a string for sql
	 *
	 * @param $node	the node which ascendants are asked
	 * @param $as_array	boolean, do you want an array (true) or a string (false)
	 * @return array or string
	 */
	public function descendants($node, $as_array = false)
	{
		if ($as_array)
			return (isset($this->descendants[$node]) ? $this->descendants[$node] : array());
		else
			return (isset($this->descendants[$node]) ? implode(',', $this->descendants[$node]) : '');
	}

	/**
	 * Returns an array of nodes, either the whole tree, or only the nodes listed in the filter
	 *
	 * @param array $filter
	 * @return array
	 */
	public function nodes(array $filter = null)
	{
		if ($filter) return array_intersect_key($this->elements, $filter);
		else return $this->elements;
	}
}

























///////////////////////// WINDOWS COMPATIBILITY FUNCTIONS //////////////////////////////


// thanks to me at rowanlewis dot com (http://fr2.php.net/manual/en/function.fnmatch.php)

if (!function_exists('fnmatch')) {
				define('FNM_PATHNAME', 1);
				define('FNM_NOESCAPE', 2);
				define('FNM_PERIOD', 4);
				define('FNM_CASEFOLD', 16);

				function fnmatch($pattern, $string, $flags = 0) {
						return pcre_fnmatch($pattern, $string, $flags);
				}
		}

		function pcre_fnmatch($pattern, $string, $flags = 0) {
				$modifiers = null;
				$transforms = array(
						'\*'    => '.*',
						'\?'    => '.',
						'\[\!'    => '[^',
						'\['    => '[',
						'\]'    => ']',
						'\.'    => '\.',
						'\\'    => '\\\\'
				);

				// Forward slash in string must be in pattern:
				if ($flags & FNM_PATHNAME) {
						$transforms['\*'] = '[^/]*';
				}

				// Back slash should not be escaped:
				if ($flags & FNM_NOESCAPE) {
						unset($transforms['\\']);
				}

				// Perform case insensitive match:
				if ($flags & FNM_CASEFOLD) {
						$modifiers .= 'i';
				}

				// Period at start must be the same as pattern:
				if ($flags & FNM_PERIOD) {
						if (strpos($string, '.') === 0 && strpos($pattern, '.') !== 0) return false;
				}

				$pattern = '#^'
						. strtr(preg_quote($pattern, '#'), $transforms)
						. '$#'
						. $modifiers;

				return (boolean)preg_match($pattern, $string);
		}





